tangoSession = setRefClass("tangoSession",
  fields = list(
    homepath = "character",
    connections = "list",
    workspaces = "list",
    currentWorkspace = "character",
    configuration = "list",
    data = "list"
  ),
  methods = list(
    initFields = function(){
      .self$createEmptyConfiguration()
      .self$initConnections()
      .self$initWorkspaces()
    },
    createEmptyConfiguration = function(){
      configuration <<- list(
        homepath = homepath,
        connections = list(),
        workspaces = list()
      )
    },
    addConnection = function(connectionName=NULL,host='localhost',username='',password='',auth=FALSE,db = ''){
      if(is.null(connectionName)) connectionName = host
      configuration[["connections"]][[connectionName]] <<- list(
        host = host,
        username = username,
        password = password,
        auth = auth,
        db = db
      )
      .self$initConnection(connectionName)
    },
    initConnections = function(){
      for(connectionName in names(configuration[["connections"]])){
        .self$initConnection(connectionName)
      }
    },
    initConnection = function(connectionName){
      c = configuration[["connections"]][[connectionName]]
      if(is.null(c[['db']])) c[['db']] = ''
      connections[[connectionName]] <<- tangoConnection$new(
        host = c[['host']],
        username = c[['username']],
        password = c[['password']],
        auth = c[['auth']],
        db = c[['db']]
      )
      connections[[connectionName]]$initFields()
      print('All those projects have been found in this connection:')
      print(names(connections[[connectionName]]$projects))
    },
    addEmptyWorkspace = function(workspaceName){
      configuration[['workspaces']][[workspaceName]] <<- list()
      workspaces[[workspaceName]] <<- list()
    },
    addConnectionToWorkspace = function(workspaceName,connectionName){
      configuration[['workspaces']][[workspaceName]][[connectionName]] <<- list()
      workspaces[[workspaceName]][[connectionName]] <<- list()
    },
    addProjectToWorkspace = function(workspaceName,connectionName,projectName){
      configuration[['workspaces']][[workspaceName]][[connectionName]][[projectName]] <<- c()
      workspaces[[workspaceName]][[connectionName]][[projectName]] <<- list()
    },
    listExperiments = function(workspaceName = 'default',connectionName = 'localhost',projectName){
      return(names(connections[[connectionName]]$projects[[projectName]]))
    },
    addExperimentsToWorkspace = function(workspaceName,connectionName,projectName,experimentsNames){
      configuration[['workspaces']][[workspaceName]][[connectionName]][[projectName]] <<- c(configuration[['workspaces']][[workspaceName]][[connectionName]][[projectName]],experimentsNames)
      xps = c()
      for(experimentName in experimentsNames){
        xp = connections[[connectionName]][['projects']][[projectName]][[experimentName]]
        if(!is.null(xp)){
          xp$fetchData()
          workspaces[[workspaceName]][[connectionName]][[projectName]][[experimentName]] <<- xp
          xps = c(xps,xp)
        }else{
          warning('The experiment does not exists, some project name or experiment name must be wrong')
        }
      }
      return(xps)
    },
    addExperiment = function(workspaceName='default',connectionName=NULL,host='localhost',username='',password='',auth=FALSE,db = '',projectName,experimentName){
      if(is.null(connectionName)) connectionName = host
      .self$addConnection(
        connectionName=connectionName,
        host=host              
      )
      if(is.null(.self$workspaces[[workspaceName]])){
        .self$addEmptyWorkspace(
          workspaceName=workspaceName
        )
      }
      if(is.null(.self$workspaces[[workspaceName]][[connectionName]])){
        .self$addConnectionToWorkspace(
          workspaceName=workspaceName,
          connectionName=connectionName
        )
      }
      if(is.null(.self$workspaces[[workspaceName]][[connectionName]][[projectName]])){
        .self$addProjectToWorkspace(
          workspaceName=workspaceName,
          connectionName=connectionName,
          projectName=projectName
        )
      }
      xp = .self$addExperimentsToWorkspace(
        workspaceName=workspaceName,
        connectionName=connectionName,
        projectName=projectName,
        experimentsNames=experimentName
      )
      return(xp[[1]])
    },
    initWorkspaces = function(){
      for(workspaceName in names(configuration[["workspaces"]])){
        .self$addEmptyWorkspace(workspaceName)
        .self$initWorkspace(workspaceName)
      }
    },
    initWorkspace = function(workspaceName){
      for(connectionName in names(workspaces[[workspaceName]])){
        .self$addConnectionToWorkspace(workspaceName,connectionName)
        for(projectName in names(workspaces[[workspaceName]][[connectionName]])){
          .self$addProjectToWorkspace(workspaceName,connectionName,projectName)
          .self$addExperimentsToWorkspace(workspaceName,connectionName,projectName,workspaces[[workspaceName]][[connectionName]][[projectName]])
        }
      }
    },
    extractSelections = function(df,selectionName,description = NULL,save=TRUE){
      l = list()
      if('host.name' %in% names(df)){
        splitDfByConnection = split(df, f=df$host.name)
        for(hostName in names(splitDfByConnection)){
          m = list()
          hostDf = splitDfByConnection[[hostName]]
          l[[hostName]] = connections[[hostName]]$extractSelections(hostDf,selectionName=selectionName,description=description,save=save)
        }
      }
      return(l)
    },
    extractSelectionsFromWorkFlow = function(workflow,save=TRUE){
      subsets = workflow$getSubsets()
      l = list()
      for(dfName in names(subsets)){
        l[[dfName]] = list()
        for(viewName in names(subsets[[dfName]])){
          if(nrow(subsets[[dfName]][[viewName]])!=0){
            l[[dfName]][[viewName]] = .self$extractSelections(subsets[[dfName]][[viewName]],paste(dfName,viewName,sep='_'),paste(viewName,dfName,sep=' applied to dataframe '),save=save)
          }
        }
      }
      return(l)
    },
    mergeDataframeList = function(df.list.list, name, col.name="name", list=NULL) {
      require(plyr)
      df.list = list()
      if (is.null(list)) list = 1:length(df.list.list)
      lev = c()
      for (i in list) {
        n=names(df.list.list)[i]
        lev = c(lev, n)
        df.list[[n]]=df.list.list[[n]][[name]]
        df.list[[n]][[col.name]]=n
      }
      res = rbind.fill(df.list)
      res[, col.name] = factor(res[,col.name], levels = lev) # ordering
      return(res)
  }
  )
)